"Commands for the TurboGears command line tool."
import optparse
import sys
import os
import os.path
import glob

import pkg_resources

import turbogears
from turbogears.util import get_model, get_package_name


sys.path.insert(0, os.getcwd())

no_connection_param = ["help", "list"]
no_model_param = ["help"]

def silent_os_remove(fname):
    """
    Tries to remove file FNAME but mutes any error that may happen.

    Returns True if file was actually removed and false otherwise
    """
    try:
        os.remove(fname)
        return True
    except os.error:
        pass
    return False

def get_project_config():
    """Tries to select appropriate project configuration file."""

    config = None
    if os.path.exists("setup.py"):
        config = "dev.cfg"
    else:
        config = "prod.cfg"
    return config

def load_project_config(config=None):
    """Tries to update the config, loading project settings from the config
    file specified.  If config is C{None} uses L{get_project_config} to locate
    one.
    """
    if config is None:
        config = get_project_config()
    if not os.path.isfile(config):
        print 'config file %s not found or is not a file.' % config
        return
    package = get_package_name()
    turbogears.update_config(configfile=config,
        modulename = package + ".config")

class CommandWithDB(object):
    "Base class for commands that need to use the database"
    config = None

    def __init__(self, version):
        pass

    def find_config(self):
        """Chooses the config file, trying to guess whether this is a
        development or installed project."""
        load_project_config(self.config)
        self.dburi = turbogears.config.get("sqlobject.dburi", None)
        if self.dburi and self.dburi.startswith("notrans_"):
            self.dburi = self.dburi[8:]

class SQL(CommandWithDB):
    """Wrapper command for sqlobject-admin.

    This automatically supplies sqlobject-admin with the database that
    is found in the config file. Will also supply the model module as
    appropriate."""

    desc = "Run the SQLObject manager"

    def __init__(self, version):
        if len(sys.argv) == 1 or sys.argv[1][0] == "-":
            parser = optparse.OptionParser(
                usage="%prog sql [turbogears options] [sqlobject command] " \
                      "[sqlobject options]\n" \
                      "HINT: '%prog sql help' will list the sqlobject " \
                      "commands",
                version="%prog " + version)
            parser.add_option("-c", "--config", help="config file",
                              dest="config")
            (options, args) = parser.parse_args(sys.argv[1:3])

            if not options.config:
                parser.error("Please provide a valid option or sqlobject command.")
            self.config = options.config
            # get rid of our config option
            if not args:
                del sys.argv[1:3]
            else:
                del sys.argv[1]

        self.find_config()

    def run(self):
        "Executes the sqlobject-admin code."
        if not "--egg" in sys.argv and not turbogears.util.get_project_name():
            print "this don't look like a turbogears project"
            return
        else:
            sqlobjcommand = sys.argv[1]

            if sqlobjcommand not in no_connection_param:
                if not self.dburi:
                    print """Database URI not specified in the config file (%s).
        Please be sure it's on the command line.""" % self.config
                else:
                    print "Using database URI %s" % self.dburi
                    sys.argv.insert(2, self.dburi)
                    sys.argv.insert(2, "-c")

            if sqlobjcommand not in no_model_param:
                if not "--egg" in sys.argv:
                    eggname = glob.glob("*.egg-info")
                    if not eggname or not \
                        os.path.exists(os.path.join(eggname[0], "sqlobject.txt")):
                        eggname = self.fix_egginfo(eggname)
                    eggname = eggname[0].replace(".egg-info", "")
                    if not "." in sys.path:
                        sys.path.append(".")
                        pkg_resources.working_set.add_entry(".")
                    sys.argv.insert(2, eggname)
                    sys.argv.insert(2, "--egg")

            from sqlobject.manager import command
            command.the_runner.run(sys.argv)

    def fix_egginfo(self, eggname):
        print """
This project seems incomplete. In order to use the sqlobject commands
without manually specifying a model, there needs to be an
egg-info directory with an appropriate sqlobject.txt file.

I can fix this automatically. Would you like me to?
"""
        dofix = raw_input("Enter [y] or n: ")
        if not dofix or dofix.lower()[0] == 'y':
            oldargs = sys.argv
            sys.argv = ["setup.py", "egg_info"]
            import imp
            imp.load_module("setup", *imp.find_module("setup", ["."]))
            sys.argv = oldargs

            import setuptools
            package = setuptools.find_packages()[0]
            eggname = glob.glob("*.egg-info")
            sqlobjectmeta = open(os.path.join(eggname[0], "sqlobject.txt"), "w")
            sqlobjectmeta.write("""db_module=%(package)s.model
history_dir=$base/%(package)s/sqlobject-history
""" % dict(package=package))
        else:
            sys.exit(0)
        return eggname


class Shell(CommandWithDB):
    """Convenient version of the Python interactive shell.
    This shell attempts to locate your configuration file and model module
    so that it can import everything from your model and make it available
    in the Python shell namespace."""

    desc = "Start a Python prompt with your database available"

    def run(self):
        "Run the shell"
        if not turbogears.util.get_project_name():
            print "this don't look like a turbogears project"
            return
        else:
            self.find_config()

            mod = get_model()
            if mod:
                locals = mod.__dict__
            else:
                locals = dict(__name__="tg-admin")
            try:
                # try to use IPython if possible
                import IPython
                shell = IPython.Shell.IPShell(user_ns=locals)
                shell.mainloop()
            except ImportError:
                import code
                code.interact(local=locals)

class ToolboxCommand(CommandWithDB):

    desc = "Launch the TurboGears Toolbox"

    def __init__(self, version):
        self.hostlist = ['127.0.0.1','::1']

        parser = optparse.OptionParser(
            usage="%prog toolbox [-o]", version="%prog " + version)
        parser.add_option("-n", "--no-open",
                 help="don't open browser automatically",
                 dest="noopen", action="store_true",
                 default=False)
        parser.add_option("-c", "--add-client",
                help="allow the client ip address specified to connect to catwalk (Can be specified more than once)",
                dest="host", action="append", default=None)
        (options, args) = parser.parse_args(sys.argv[1:])
        self.noopen = options.noopen
        if options.host:
            self.hostlist = self.hostlist + options.host
        turbogears.widgets.load_widgets()


    def openbrowser(self):
        import webbrowser
        webbrowser.open("http://localhost:7654")

    def run(self):
        self.find_config()
        from turbogears.toolbox.catwalk import CatWalk
        import cherrypy
        from turbogears import toolbox
        root = toolbox.Toolbox()
        model = get_model()
        if model:
            root.catwalk = CatWalk(model, self.hostlist)
        cherrypy.root = root

        turbogears.config.update({"global" : {
            "server.socket_port" : 7654,
            "server.environment" : "development",
            "server.log_to_screen" : True,
            "autoreload.on" : False,
            "server.package" : "turbogears.toolbox",
            "log_debug_info_filter.on" : False
            }})
        if not self.noopen:
            cherrypy.server.start_with_callback(self.openbrowser)
        else:
            print "Running server on port 7654."
            cherrypy.server.start()


commands = None

def main():
    "Main command runner. Manages the primary command line arguments."
    # add commands defined by entrypoints
    commands = {}
    for entrypoint in pkg_resources.iter_entry_points("turbogears.command"):
        command = entrypoint.load()
        commands[entrypoint.name] = (command.desc, entrypoint)
    if len(sys.argv) == 1 or sys.argv[1] == "--help-commands" or \
        sys.argv[1] == "--help" or not commands.has_key(sys.argv[1]):
        print """
TurboGears %s command line interface

Usage: %s [command] [options]

Available commands:""" % (turbogears.__version__, sys.argv[0])
        longest = max([len(key) for key in commands.keys()])
        format = "%" + str(longest) + "s  %s"
        commandlist = commands.keys()
        commandlist.sort()
        for key in commandlist:
            print format % (key, commands[key][0])
        sys.exit(0)
    commandname = sys.argv[1]
    del sys.argv[1]
    command = commands[commandname][1]
    command = command.load()(turbogears.__version__)
    command.run()



__all__ = ["main"]
