"""Provides convenient access to an SQLObject-managed database."""

import sys,re

import sqlobject
from sqlobject.dbconnection import ConnectionHub, Transaction, TheURIOpener
import cherrypy
from cherrypy.filters.basefilter import BaseFilter
import turbogears

# Provide support for sqlalchemy
try:
    import sqlalchemy
    from sqlalchemy.ext.proxy import AutoConnectEngine

    # declare the PackageEngine class
    class PackageEngine(AutoConnectEngine):
        '''
        An AutoConnectEngine that looks for the dburi based on a package name.
        '''
        def __init__(self, package_name, *args, **kwargs):
            def lookup_dburi():
                uri= turbogears.config.get( "%s.sqlalchemy.dburi" % package_name,
                                          None )
                return uri or turbogears.config.get( "sqlalchemy.dburi", None )

            AutoConnectEngine.__init__(self, lookup_dburi, *args, **kwargs)

except ImportError:
    sqlalchemy= None


try:
    set
except NameError:
    from sets import Set as set

hub_registry = set()

class AutoConnectHub(ConnectionHub):
    """Connects to the database once per thread. The AutoConnectHub also
    provides convenient methods for managing transactions."""
    uri = None
    params = {}

    def __init__(self, uri=None, supports_transactions=True):
        if not uri:
            uri = turbogears.config.get("sqlobject.dburi")
        if sys.platform == 'win32' and uri.startswith('sqlite'):
            # fix drive letter path
            uri = re.sub('///(\w):', '///\\1|', uri)
        self.uri = uri
        self.supports_transactions = supports_transactions
        hub_registry.add(self)
        ConnectionHub.__init__(self)

    def getConnection(self):
        try:
            conn = self.threadingLocal.connection
            return self.begin(conn)
        except AttributeError:
            if self.uri:
                conn = sqlobject.connectionForURI(self.uri)
                # the following line effectively turns off the DBAPI connection
                # cache. We're already holding on to a connection per thread,
                # and the cache causes problems with sqlite.
                if self.uri.startswith("sqlite"):
                    TheURIOpener.cachedURIs = {}
                self.threadingLocal.connection = conn
                return self.begin(conn)
            raise AttributeError(
                "No connection has been defined for this thread "
                "or process")

    def reset(self):
        """Used for testing purposes. This drops all of the connections
        that are being held."""
        self.threadingLocal = threading_local()

    def begin(self, conn=None):
        "Starts a transaction."
        if not self.supports_transactions:
            return conn
        if not conn:
            conn = self.getConnection()
        if isinstance(conn, Transaction):
            if conn._obsolete:
                conn.begin()
            return conn
        self.threadingLocal.old_conn = conn
        trans = conn.transaction()
        self.threadingLocal.connection = trans
        return trans

    def commit(self):
        "Commits the current transaction."
        if not self.supports_transactions:
            return
        try:
            conn = self.threadingLocal.connection
        except AttributeError:
            return
        if isinstance(conn, Transaction):
            self.threadingLocal.connection.commit()

    def rollback(self):
        "Rolls back the current transaction."
        if not self.supports_transactions:
            return
        try:
            conn = self.threadingLocal.connection
        except AttributeError:
            return
        if isinstance(conn, Transaction) and not conn._obsolete:
            self.threadingLocal.connection.rollback()

    def end(self):
        "Ends the transaction, returning to a standard connection."
        if not self.supports_transactions:
            return
        try:
            conn = self.threadingLocal.connection
        except AttributeError:
            return
        if not isinstance(conn, Transaction):
            return
        if not conn._obsolete:
            conn.rollback()
        self.threadingLocal.connection = self.threadingLocal.old_conn
        del self.threadingLocal.old_conn
        self.threadingLocal.connection.cache.clear()

# This dictionary stores the AutoConnectHubs used for each
# connection URI
_hubs = dict()

class PackageHub(object):
    """Transparently proxies to an AutoConnectHub for the URI
    that is appropriate for this package. A package URI is
    configured via "packagename.dburi" in the global CherryPy
    settings. If there is no package DB URI configured, the
    default (provided by "sqlobject.dburi") is used.

    The hub is not instantiated until an attempt is made to
    use the database.
    """
    def __init__(self, packagename):
        self.packagename = packagename
        self.hub = None

    def __get__(self, obj, type):
        if not self.hub:
            self.set_hub()
        return self.hub.__get__(obj, type)

    def __set__(self, obj, type):
        if not self.hub:
            self.set_hub()
        return self.hub.__set__(obj, type)

    def __getattr__(self, name):
        if not self.hub:
            self.set_hub()
        return getattr(self.hub, name)

    def set_hub(self):
        dburi = turbogears.config.get("%s.dburi" % self.packagename, None)
        if not dburi:
            dburi = turbogears.config.get("sqlobject.dburi", None)
        if not dburi:
            raise KeyError, "No database configuration found!"
        if dburi.startswith("notrans_"):
            dburi = dburi[8:]
            trans = False
        else:
            trans = True
        hub = _hubs.get(dburi, None)
        if not hub:
            hub = AutoConnectHub(dburi, supports_transactions=trans)
            _hubs[dburi] = hub
        self.hub = hub

def set_db_uri(dburi, package=None):
    """Sets the database URI to use either globally or for a specific
    package. Note that once the database is accessed, calling
    setDBUri will have no effect.

    @param dburi: database URI to use
    @param package: package name this applies to, or None to set the default.
    """
    if package:
        turbogears.config.update({'global':
            {"%s.dburi" % package : dburi}
        })
    else:
        turbogears.config.update({'global':
            {"sqlobject.dburi" : dburi}
        })

def commit_all():
    "Commits the Transactions in all registered hubs (for this thread)"
    for hub in hub_registry:
        hub.commit()

def rollback_all():
    "Rolls back the Transactions in all registered hubs (for this thread)"
    for hub in hub_registry:
        hub.rollback()

def end_all():
    "Ends the Transactions in all registered hubs (for this thread)"
    for hub in hub_registry:
        hub.end()

def run_with_transaction(func, *args, **kw):
    try:
        try:
            if sqlalchemy:
                # flush any cached objects
                sqlalchemy.objectstore.clear()
            retval = func(*args, **kw)
            if sqlalchemy:
                sqlalchemy.objectstore.commit()
            commit_all()
            return retval
        except cherrypy.HTTPRedirect:
            commit_all()
            if sqlalchemy:
                sqlalchemy.objectstore.commit()
            raise
        except cherrypy.InternalRedirect:
            commit_all()
            if sqlalchemy:
                sqlalchemy.objectstore.commit()
            raise
        except:
            # No need to "rollback" the sqlalchemy unit of work, because nothing
            # has hit the db yet.
            rollback_all()
            raise
    finally:
        end_all()

def so_to_dict(sqlobj):
    "Converts SQLObject to a dictionary based on columns"
    d = {}
    if sqlobj == None:
        # stops recursion
        return d
    for name in sqlobj.sqlmeta.columns.keys():
        d[name] = getattr(sqlobj, name)
    "id must be added explicitly"
    d["id"] = sqlobj.id
    if sqlobj._inheritable:
        d.update(so_to_dict(sqlobj._parent))
        d.pop('childName')
    return d

class EndTransactionsFilter(BaseFilter):
    def on_end_resource(self):
        end_all()

__all__ = ["PackageHub", "AutoConnectHub", "set_db_uri",
           "commit_all", "rollback_all", "end_all", "so_to_dict",
           "EndTransactionsFilter"]

if sqlalchemy:
    __all__.append( "PackageEngine" )
