import cherrypy
import sha
import md5
import random
import turbogears
from turbojson.jsonify import *
from datetime import *
from sqlalchemy import *
from sqlalchemy.ext.activemapper import *

import logging
log = logging.getLogger("turbogears.identity.saprovider")

from turbogears import identity

__engine__= turbogears.database.PackageEngine( 'qotd' )

try:
    set, frozenset
except NameError:
    from sets import Set as set, ImmutableSet as frozenset

# Global class references -- these will be set when the Provider is initialised.
user_class= None
group_class= None
permission_class= None

class SqlAlchemyIdentity(object):
    def __init__(self, visit_key, user=None):
        if user:
            self._user= user
        self.visit_key= visit_key
    
    def _get_user(self):
        try:
            return self._user
        except AttributeError:
            # User hasn't already been set
            pass
        # Attempt to load the user. After this code executes, there *WILL* be
        # a _user attribute, even if the value is None.
        visit= TG_VisitIdentity.get_by( visit_key= self.visit_key )
        if not visit:
            self._user= None
            return None
        self._user= user_class.get( visit.user_id )
        return self._user
    user= property(_get_user)
    
    def _get_user_name(self):
        if not self.user:
            return None
        return self.user.user_name
    user_name= property(_get_user_name)

    def _get_anonymous(self):
        return not self.user
    anonymous= property(_get_anonymous)
    
    def _get_permissions(self):
        try:
            return self._permissions
        except AttributeError:
            # Permissions haven't been computed yet
            pass
        if not self.user:
            self._permissions= frozenset()
        else:
            self._permissions= frozenset([p.permission_name for p in self.user.permissions])
        return self._permissions
    permissions= property(_get_permissions)
    
    def _get_groups(self):
        try:
            return self._groups
        except AttributeError:
            # Groups haven't been computed yet
            pass
        if not self.user:
            self._groups= frozenset()
        else:
            self._groups= frozenset([g.group_name for g in self.user.groups])
        return self._groups
    groups= property(_get_groups)

    def logout(self):
        '''
        Remove the link between this identity and the visit.
        '''
        if not self.visit_key:
            return
        try:
            visit= TG_VisitIdentity.get_by(visit_key=self.visit_key)
            objectstore.delete( visit )
            # Clear the current identity
            anon= SqlAlchemyIdentity(None,None)
            anon.anonymous= True
            identity.set_current_identity( anon )
        except:
            pass


def _load_class(dottedpath):
    '''
    Loads a class from a module in dotted-path notation.
    Eg: load_class("package.module.class").

    Based on recipe 16.3 from "Python Cookbook, 2ed., by Alex Martelli,
    Anna Martelli Ravenscroft, and David Ascher (O'Reilly Media, 2005)
    0-596-00797-3"
    '''
    splitted_path= dottedpath.split('.')
    modulename= '.'.join(splitted_path[:-1])
    classname= splitted_path[-1]
    try:
        module = __import__(modulename, globals(), locals(), [classname])
        log.info("Succesfully loaded \"%s\"" % dottedpath)
    except ImportError:
        return None
    return getattr(module, classname)

    
class SqlAlchemyIdentityProvider(object):
    '''
    IdentityProvider that uses a model from a database (via SQLObject).
    '''
    
    def __init__(self):
        super(SqlAlchemyIdentityProvider, self).__init__()
        get=cherrypy.config.get
        
        global user_class, group_class, permission_class
        
        user_class_path= get( "identity.saprovider.model.user", 
                              __name__ + ".TG_User" )
        user_class= _load_class(user_class_path)
        group_class_path= get( "identity.saprovider.model.group",
                                __name__ + ".TG_Group" )
        group_class= _load_class(group_class_path)
        permission_class_path= get( "identity.saprovider.model.permission",
                                    __name__ + ".TG_Permission" )
        permission_class= _load_class(permission_class_path)
            
        # Default encryption algorithm is to use plain text passwords
        algorithm= get( "identity.saprovider.encryption_algorithm", None )
        if "md5"==algorithm:
            self.encrypt_password= lambda pw: md5.new(pw).hexdigest()
        elif "sha1"==algorithm:
            self.encrypt_password= lambda pw: sha.new(pw).hexdigest()
        else:
            self.encrypt_password= lambda pw: pw

    def create_provider_model( self ):
        # create the database tables
        try:
            tg_user_group.create()
            tg_group_permission.create()
            TG_VisitIdentity.table.create()
            user_class.table.create()
            group_class.table.create()
            permission_class.table.create()
            objectstore.commit()
        except SQLError:
            return
        except KeyError:
            log.warning( "No database is configured: SqlAlchemyIdentityProvider is disabled." )
            return

    def validate_identity( self, user_name, password, visit_key ):
        '''
        Look up the identity represented by user_name and determine whether the
        password is correct.
        
        Must return either None if the credentials weren't valid or an object
        with the following properties:
            user_name: original user name
            user: a provider dependant object (TG_User or similar)
            groups: a set of group IDs
            permissions: a set of permission IDs
        '''
        user= user_class.get_by( user_name=user_name )
        if not user:
            log.warning( "No such user: %s", user_name )
            return None
        if user.password!=self.encrypt_password(password):
            log.info( "Passwords don't match for user: %s", user_name )
            return None
        
        log.info( "associating user (%s) with visit (%s)", user.user_name, 
                  visit_key )
        # Link the user to the visit
        link= TG_VisitIdentity.get_by( visit_key=visit_key )
        if not link:
            link= TG_VisitIdentity( visit_key=visit_key, user_id=user.user_id )
        else:
            link.user_id= user.user_id
        link.commit()
        return SqlAlchemyIdentity( visit_key, user )
            
    def load_identity( self, visit_key ):
        '''
        Lookup the principal represented by user_name. Return None if there is no
        principal for the given user ID.
        
        Must return an object with the following properties:
            user_name: original user name
            user: a provider dependant object (TG_User or similar)
            groups: a set of group IDs
            permissions: a set of permission IDs
        '''
        return SqlAlchemyIdentity( visit_key )
    
    def anonymous_identity( self ):
        '''
        Must return an object with the following properties:
            user_name: original user name
            user: a provider dependant object (TG_User or similar)
            groups: a set of group IDs
            permissions: a set of permission IDs
        '''
        return SqlAlchemyIdentity( None )
        

# tables for SQLAlchemy identity
tg_user_group= Table( "tg_user_group", __engine__, 
                      Column( "user_id", Integer,
                              ForeignKey("tg_user.user_id"),
                              primary_key=True ),
                      Column( "group_id", Integer,
                              ForeignKey("tg_group.group_id"),
                              primary_key=True ) )
                      
tg_group_permission= Table( "tg_group_permission", __engine__,
                            Column( "group_id", Integer,
                                    ForeignKey("tg_group.group_id"),
                                    primary_key=True ),
                            Column( "permission_id", Integer,
                                    ForeignKey("tg_permission.permission_id"),
                                    primary_key=True ) )


class TG_VisitIdentity(ActiveMapper):
    class mapping:
        __table__="tg_visit_identity"
        visit_key= column( String, # foreign_key="tg_visit.visit_key",
                          primary_key=True )
        user_id= column( Integer, foreign_key="tg_user.user_id", index=True )


class TG_Group(ActiveMapper):
    '''
    An ultra-simple group definition.
    '''
    class mapping:
        __table__="tg_group"
        group_id= column( Integer, primary_key=True )
        group_name= column( Unicode(16), unique=True )
        display_name= column( Unicode(255) )
        created= column( DateTime, default=datetime.now )

        users= many_to_many( "TG_User", tg_user_group, backref="groups" )
        permissions= many_to_many( "TG_Permission", tg_group_permission,
                                   backref="groups" )


class TG_User(ActiveMapper):
    '''
    Reasonably basic User definition. Probably would want additional attributes.
    '''
    class mapping:
        __table__="tg_user"
        user_id= column( Integer, primary_key=True )
        user_name= column( Unicode(16), unique=True )
        email_address= column( Unicode(255), unique=True )
        display_name= column( Unicode(255) )
        password= column( Unicode(40) )
        created= column( DateTime, default=datetime.now )

        groups= many_to_many( "TG_Group", tg_user_group, backref="users" )

    def _get_permissions( self ):
        perms= set()
        for g in self.groups:
            perms= perms | set(g.permissions)
        return perms
    permissions= property(_get_permissions)
        

class TG_Permission(ActiveMapper):
    class mapping:
        __table__="tg_permission"
        permission_id= column( Integer, primary_key=True )
        permission_name= column( Unicode(16), unique=True )
        description= column( Unicode(255) )
        
        groups= many_to_many( "TG_Group", tg_group_permission,
                              backref="permmissions" )
