import unittest
import os
import tempfile
from datetime import datetime

import kid

import turbogears

locale_dir = os.path.join(os.path.dirname(__file__), 'locale')

dt = datetime(2005, 2, 5) # Saturday, 5rd Feb 2005
def get_test_locale():
    return turbogears.config.get('i18n.defaultLocale', 'en')

turbogears.config.update({
    'i18n.localeDir':locale_dir,
    'i18n.domain':'messages',
    'i18n.defaultLocale':'en',
    'i18n.getLocale':get_test_locale,
    'i18n.run_template_filter':False,
    'sqlobject.dburi':"sqlite:///" + tempfile.gettempdir().replace(':\\','|/')  + "/testlocaledb"
})

from turbogears.i18n import *
from turbogears.i18n import sogettext
from turbogears.i18n.tg_gettext import tg_gettext

# Create database data from messages files

sogettext.create_so_catalog(["en","fi"], "messages")

class TestI18n(unittest.TestCase):
    
    def test_gettext(self):

        assert unicode(gettext("Welcome")) == "Welcome"
        assert unicode(gettext("Welcome", "en")) == "Welcome"
        assert unicode(gettext("Welcome", "fi")) == "Tervetuloa"
        assert unicode(gettext("Welcome", "fi_FI")) == "Tervetuloa"
    
    def test_invalid_domain(self):
    
        assert gettext("Welcome", "fi", "fubar") != "Tervetuloa"
            
    def test_is_unsupported_locale(self):
    
        assert is_locale_supported("en") == True
        assert is_locale_supported("en-gb") == False
        assert is_locale_supported("de") == False
        assert is_locale_supported("en", "fubar") == False
        
    def test_gettext_unsupported_locale(self):
    
        assert unicode(gettext("Welcome", "en-gb")) == "Welcome"
        assert unicode(gettext("Welcome", "de")) == "Welcome"
    
    def test_get_accept_languages(self):
    
        assert get_accept_languages("da, en-gb;q=0.8, en;q=0.7") == ["da", "en_GB", "en"]
        assert get_accept_languages("da;q=0.6, en-gb;q=1.0, en;q=0.7") == ["en_GB", "en", "da"]
    
    def test_ngettext(self): 
        
        assert ngettext("You have %i apple", "You have %i apples", 1) %1 == "You have 1 apple"
        assert ngettext("You have %i apple", "You have %i apples", 1, "fi") %1 == "Sinulla on 1 omena"
        assert ngettext("You have %i apple", "You have %i apples", 3) %3 == "You have 3 apples"
        assert ngettext("You have %i apple", "You have %i apples", 3, "fi") %3== "Sinulla on 3 omenaa"
    
    def test_match_template(self):
    
        template = """
        <?python
        from turbogears.i18n import translate
        ?>
        <html xmlns:py="http://purl.org/kid/ns#">
            <translate py:match="item.attrib.has_key('lang')" py:replace="translate(item, 'lang')"/>
            <body lang="">
                Welcome
                <p>Welcome</p>  
                <p lang="en">Welcome</p>
                <p lang="fi">Welcome</p>
            </body>
        </html>
        """
        
        t = kid.Template(source = template)
        output = t.serialize()
        print output
        assert '<p lang="en">Welcome</p>' in output
        assert '<p lang="fi">Tervetuloa</p>' in output

    def test_i18n_filter(self):

        template = """
        <html xmlns:py="http://purl.org/kid/ns#">
            <body>
                Welcome, <em>guest</em>!
                <p>Welcome</p>  
                <p lang="en">Welcome</p>
                <p lang="fi">Welcome<a href="" lang="de">Welcome</a></p>
            </body>
        </html>"""
        
        t = kid.Template(source = template)
        t._filters+=[i18n_filter]
        output = t.serialize()
        print output

     
    def test_get_countries(self):
    
        print get_countries()
        print get_countries("eo")
        print get_countries("en_US")
        
    def test_get_languages(self):
    
        print get_languages()
        print get_languages("eo")
        print get_countries("en_US")

    def test_get_month_names(self):

        print get_month_names()
        print get_abbr_month_names()
    
        print get_month_names("eo")
        print get_abbr_month_names("eo")

        print get_month_names("en_US")
        print get_abbr_month_names("en_US")

    def test_get_weekday_names(self):

        print get_weekday_names()
        print get_abbr_weekday_names()
    
        print get_weekday_names("eo")
        print get_abbr_weekday_names("eo")

        print get_weekday_names("en_US")
        print get_abbr_weekday_names("en_US")
    
    def test_format_numbers(self):

        assert format_number(50, "en_US")=="50"
        assert format_number(50, "de")=="50"

        assert format_number(500, "en_US")=="500"
        assert format_number(500, "de")=="500"

        assert format_number(5000, "en_US")=="5,000"
        assert format_number(5000, "de")=="5.000"

        assert format_number(50000, "en_US")=="50,000"
        assert format_number(50000, "de")=="50.000"

        assert format_number(500000, "en_US")=="500,000"
        assert format_number(500000, "de")=="500.000"
            
        assert format_number(5000000, "en_US")=="5,000,000"
        assert format_number(5000000, "de")=="5.000.000"
    
        assert format_decimal(5000000.78781, 5, "en_US")=="5,000,000.78781"
        assert format_decimal(5000000.78781, 5, "de")=="5.000.000,78781"

        assert format_currency(5000000.78781, "en_US")=="5,000,000.79"
        assert format_currency(5000000.78781, "de")=="5.000.000,79"
    
    def test_format_date(self):
                
        assert format_date(dt, "long", "eo") == u"2005-februaro-05"
        assert format_date(dt, "medium", "eo") == u"2005-feb-05"
        assert format_date(dt, "short", "eo") == u"05-02-05"
        assert format_date(dt, "full", "eo") == u"sabato, 05-a de februaro 2005"
    
    def test_parse_numbers(self):
    
        assert parse_number("5,000,000", "en_US") == 5000000
        assert parse_number("5.000.000", "de") == 5000000
        assert parse_decimal("5,000,000.78781", "en_US") == 5000000.78781
        assert parse_decimal("5.000.000,78781", "de") == 5000000.78781
    
        try:
            parse_number("5.000.000", "en_US")
            parse_number("5.000.000,78781", "en_US")
            assert False
        except ValueError:
            assert True            
            
    def test_diff_locale_format_date(self):
        
        assert format_date(dt, "short", "en_US") == "02/05/05"
        assert format_date(dt, "short", "en_GB") == "05/02/2005"

    def test_invalid_locale_format(self):
        
        assert get_abbr_weekday_names("fubar")==[]
        
    def test_so_gettext(self):
    
        # empty thread data
        
        turbogears.config.update({"i18n.gettext":sogettext.so_gettext})
        self.test_gettext()
        self.test_ngettext()
        self.test_match_template()
        self.test_i18n_filter()
        self.test_invalid_domain()
        turbogears.config.update({"i18n.gettext":tg_gettext})
    
if __name__=="__main__":
    
    unittest.main()
