"""Graphical user interface for managing TurboGears projects"""

from turbogears import controllers
from turbogears.util import setlike
import turbogears
import cherrypy
import shell
import admi18n
import designer

class Info(controllers.RootController):
    """TurboGears System Information.
       Lists your TurboGears packages and version information
    """
    __label__ ="System Info"
    __version__ = "0.1"
    __author__ = "Ronald Jaramillo"
    __email__ = "ronald@checkandshare.com"
    __copyright__ = "Copyright 2005 Ronald Jaramillo"
    __license__ = "MIT"

    def index(self):
        from turbogears.command.info import retrieve_info
        packages, plugins = retrieve_info()
        return dict(packages=packages, plugins=plugins)
    index = turbogears.expose(template='turbogears.toolbox.info')(index)

class WidgetBrowser(controllers.Controller):
    """Browse usage samples, description and source code for the available TurboGears Widgets"""
    __label__ ="Widget Browser"
    __version__ = "0.1"
    __author__ = "Kevin Dangoor"
    __email__ = "dangoor+turbogears@gmail.com"
    __copyright__ = "Copyright 2005 Kevin Dangoor"
    __license__ = "MIT"
    
    all_descs = None
    
    [turbogears.expose(template="turbogears.toolbox.widgets")]
    def index(self, name=None):
        from turbogears import widgets
        from turbogears.widgets import js_location
        all_descs = self.all_descs
        if not all_descs:
            widgets.load_widgets()
            all_descs = dict()
            for widgetdesc in widgets.all_widgets:
                wd = widgetdesc()
                all_descs[wd.full_class_name.replace(".", "_")] = wd
            self.all_descs = all_descs
        if name:
            all_descs = {name : all_descs[name]}
        desclist = list(all_descs.values())
        desclist.sort(cmp=lambda x,y: cmp(x.name, y.name))
        output = dict(descs=desclist, viewing_one=name != None)
        
        css = setlike()
        js = dict()
        for l in js_location:
            js[l] = setlike()
        
        for widgetdesc in desclist:
            if not name and widgetdesc.show_separately:
                continue
            css.add_all(widgetdesc.retrieve_css())
            
            for l in js_location:
                for script in widgetdesc.retrieve_javascript():
                    if hasattr(script, "location"):
                        js[script.location].add(script)
                    else:
                        js[js_location.head].add(script)
        output["widget_css"] = css
        for l in js_location:
            output["widget_js_%s" % str(l)] = js[l]
        
        return output
    
    def __getattr__(self, widgetname):
        try:
            return self.all_descs[widgetname]
        except:
            raise AttributeError(widgetname)

class Toolbox(controllers.RootController):
    widgets = WidgetBrowser()
    shell = shell.WebConsole()
    admi18n = admi18n.Internationalization() 
    designer = designer.Designer()
    info = Info()

    def tool_icon(self,tool):
        t = getattr(self,tool)
        icon = getattr(t,'__icon__','')
        if icon: return icon
        return '/tg_static/images/%s.png'% tool

    def get_tools(self):
        tools = ['info','widgets','shell','admi18n','designer']
        if hasattr(self, 'catwalk'):tools.append('catwalk')
        toolbox = []
        for tool in tools:
            t = getattr(self,tool)
            toolbox.append(
                           {
                            'path':tool,
                            'label':getattr(t,'__label__',tool),
                            'description':getattr(t,'__doc__',''),
                            'version':getattr(t,'__version__',''),
                            'author':getattr(t,'__author__',''),
                            'email':getattr(t,'__email__',''),
                            'copyright':getattr(t,'__copyright__',''),
                            'license':getattr(t,'__license__',''),
                            'icon':self.tool_icon(tool)
                           }
                          )
        return toolbox
    
    def arrange_in_pairs(self,tools):
        p = [[],[]]
        for idx,tool in enumerate(tools): p[idx%2].append(tool)
        pairs = zip(p[0],p[1]) 
        if len(p[0]) > len(p[1]): pairs.append([p[0][-1],None])
        if len(p[0]) < len(p[1]): pairs.append([p[1][-1],None])
        return pairs
        
    def index(self):
        return dict(toolbox = self.arrange_in_pairs(self.get_tools()))
    index = turbogears.expose(template="turbogears.toolbox.main")(index)
