"""Convenient validators and converters for data coming in from the web.

This module also imports everything from formencode.validators, so all
common validation routines are available here."""

import time
from datetime import datetime
import cgi # FieldStorageUploadConverter

import simplejson
from formencode.validators import *
from formencode.compound import *
from formencode.api import Invalid, NoDefault
from formencode.schema import Schema
from formencode import ForEach

from turbogears.i18n import format
from turbogears import util
from turbojson import jsonify

from formencode import validators # Needed to disambiguate the Number validator... 

class Money(FancyValidator):
    messages = {
        'badFormat': 'Invalid number format',
        'empty': 'Empty values not allowed',
    }

    def __init__(self, allow_empty = False, *args, **kwargs):
        self.allow_empty = allow_empty


    def _to_python(self, value, state):
        """ parse a string and returns a float or integer """
        if self.allow_empty and not value:
            return None
        elif not self.allow_empty and not value:
            raise Invalid(self.message('empty', state), value, state)
        else:
            try:
                return format.parse_decimal(value)
            except ValueError:
                raise Invalid(self.message('badFormat', state), value, state)

    def _from_python(self, value, state):
        """ returns a string using the correct grouping """
        if self.allow_empty and not value:
            return None
        else:
            return format.format_currency(value)


class Number(FancyValidator):

    def _to_python(self, value, state):
        """ parse a string and returns a float or integer """
        try:
            value = format.parse_decimal(value)
        except ValueError:
            pass
        nb = validators.Number()
        return nb._to_python(value, state)

    def _from_python(self, value, state):
        """ returns a string using the correct grouping """
        dec_places = util.find_precision(value)
        if dec_places > 0:
            return format.format_decimal(value, dec_places)
        else:
            return format.format_number(value)


class DateTimeConverter(FancyValidator):
    messages = {
        'badFormat': 'Invalid datetime format',
        'empty': 'Empty values not allowed',
    }

    def __init__(self, format = "%Y/%m/%d %H:%M", allow_empty=False, *args, **kwargs):
        self.format = format
        self.allow_empty = allow_empty

    def _to_python(self, value, state):
        """ parse a string and return a datetime object. """
        if self.allow_empty and not value:
            return None
        if value and isinstance(value, datetime):
            return value
        elif not self.allow_empty and not value:
            raise Invalid(self.message('empty', state), value, state)
        else:
            try:
                tpl = time.strptime(value, self.format)
            except ValueError:
                raise Invalid(self.message('badFormat', state), value, state)
            # shoudn't use time.mktime() because it can give OverflowError,
            # depending on the date (e.g. pre 1970) and underlying C library
            return datetime(year=tpl.tm_year, month=tpl.tm_mon, day=tpl.tm_mday,
                    hour=tpl.tm_hour, minute=tpl.tm_min, second=tpl.tm_sec)

    def _from_python(self, value, state):
        if self.allow_empty and not value:
            return None
        else:
            return value.strftime(self.format)

# formencode trunk already contains UnicodeString but we
# provide our own version because formencode's version may be broken
# remove this when formencode.validators.UnicodeString will be identical to ours.
class UnicodeString(String):
    encoding = 'utf-8'
    def _to_python(self, value, state):
        if value:
            if isinstance(value, unicode):
                return value
            if hasattr(value, '__unicode__'):
                return unicode(value)
            return unicode(value, self.encoding)
        return u''
    def _from_python(self, value, state):
        if hasattr(value, '__unicode__'):
            value = unicode(value)
        if isinstance(value, unicode):
            return value.encode(self.encoding)
        return str(value)

# another formencode workaround,
# see #1464357 on FE bugtracker (http://tinyurl.com/lm9ae).
# Custom version of FieldStorage validator that does not break FE schema validator.
class FieldStorageUploadConverter(FancyValidator):
    def to_python(self, value, state=None):
        if isinstance(value, cgi.FieldStorage):
            if value.filename:
                return value
            raise Invalid('invalid', value, state)
        else:
            return value 

class MultipleSelection(ForEach):
    if_missing = NoDefault
    if_empty = []

    def to_python(self, value, state=None):
        try:
            return super(MultipleSelection, self).to_python(value, state)
        except Invalid:
            raise Invalid("Please select at least a value", value, state)

class Schema(Schema):
    """ A Schema validator """
    filter_extra_fields = True
    allow_extra_fields = True
    if_key_missing = None

class JSONValidator(FancyValidator):

    def _from_python(self, value, state):
        return jsonify.encode(value)

    def _to_python(self, value, state):
        return simplejson.loads(value)

